# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
CONV_TYPE = "Conv"
BN_TYPE = "BatchNormalization"
SLICE_TYPE = "Slice"
CONCAT_TYPE = "Concat"
MAXPOOL_TYPE = "MaxPool"
AVGPOOL_TYPE = "AveragePool"
RELU_TYPE = "Relu"
ADD_TYPE = "Add"
FC_TYPE = "Gemm"
RESHAPE_TYPE = "Reshape"
GAP_TYPE = "GlobalAveragePool"
CLIP_TYPE = "Clip"
MUL_TYPE = "Mul"
DIV_TYPE = "Div"
HARDSIGMOID_TYPE = "HardSigmoid"
FLATTEN_TYPE = "Flatten"
TRANSPOSE_TYPE = "Transpose"
REDUCEMEAN_TYPE = "ReduceMean"
SPLIT_TYPE = "Split"
PAD_TYPE = "Pad"
